import hashlib
import logging
import uuid

from django.conf import settings
from django.db import transaction
from django.shortcuts import reverse
from django.utils import timezone
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.ponos.keys import check_agent_key
from arkindex.ponos.models import (
    FINAL_STATES,
    GPU,
    ActionType,
    Agent,
    Artifact,
    Farm,
    Secret,
    State,
    Task,
    Workflow,
    task_token_default,
)
from arkindex.ponos.serializer_fields import Base64Field, PublicKeyField
from arkindex.ponos.signals import task_failure
from arkindex.project.serializer_fields import EnumField

logger = logging.getLogger(__name__)


class FarmSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Farm` instance.
    """

    class Meta:
        model = Farm
        fields = (
            "id",
            "name",
        )


class GPUSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.GPU` instance for public access.
    """

    class Meta:
        model = GPU
        fields = (
            "id",
            "index",
            "name",
            "ram_total",
        )
        read_only_fields = (
            "id",
            "index",
            "name",
            "ram_total",
        )


class AgentLightSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Agent` instance for public access.
    """

    farm = FarmSerializer()
    gpus = GPUSerializer(many=True, read_only=True)

    class Meta:
        model = Agent
        fields = (
            "id",
            "farm",
            "hostname",
            "cpu_cores",
            "cpu_frequency",
            "ram_total",
            "cpu_load",
            "ram_load",
            "last_ping",
            "gpus",
        )
        read_only_fields = ("id", "last_ping")


class TaskLightSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance without logs or agent information.
    Used to list tasks inside a workflow.
    """

    state = EnumField(State)
    url = serializers.HyperlinkedIdentityField(view_name="api:task-details")
    tags = serializers.ListField(
        child=serializers.CharField(),
        allow_empty=True,
        read_only=True,
    )

    class Meta:
        model = Task
        fields = (
            "id",
            "url",
            "run",
            "depth",
            "slug",
            "state",
            "parents",
            "tags",
            "shm_size",
        )
        read_only_fields = (
            "id",
            "url",
            "run",
            "depth",
            "slug",
            "parents",
            "tags",
            "shm_size",
        )


class TaskSerializer(TaskLightSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance with logs and agent information.
    """

    logs = serializers.CharField(source="short_logs", read_only=True)
    full_log = serializers.URLField(source="s3_logs_get_url", read_only=True)
    agent = AgentLightSerializer(read_only=True)
    gpu = GPUSerializer(read_only=True)
    extra_files = serializers.DictField(read_only=True)

    class Meta(TaskLightSerializer.Meta):
        fields = TaskLightSerializer.Meta.fields + (
            "logs",
            "full_log",
            "agent",
            "gpu",
            "extra_files",
        )
        read_only_fields = TaskLightSerializer.Meta.read_only_fields + (
            "logs",
            "full_log",
            "agent",
            "gpu",
            "extra_files",
        )

    def update(self, instance: Task, validated_data) -> Task:
        """
        Perform updates on a task instance.
        If a task gets marked as :attr:`~arkindex.ponos.models.State.Completed`, this will set
        its child tasks to :attr:`~arkindex.ponos.models.State.Pending` if all of their dependencies are met.
        """
        instance = super().update(instance, validated_data)

        if instance.state not in FINAL_STATES:
            return instance

        with transaction.atomic(using="default"):
            # When a task is completed, set its children to be started next
            # We need to use the default Database here to avoid a stale read
            # as the parent instance has been updated just before
            children = list(
                instance.children.using("default").filter(state=State.Unscheduled)
            )
            if instance.state == State.Completed:
                for child in children:
                    if (
                        child.parents.using("default")
                        .exclude(state=State.Completed)
                        .exists()
                    ):
                        # This child has another parent that is not completed
                        continue
                    logging.info(f"Setting child task {child} to Pending state")
                    child.state = State.Pending
                    child.save()

            # This task has no children: this might be the last task of the workflow, so the workflow might be finished.
            # If all tasks in the current run are finished, update the completion date of the workflow.
            if (
                not children
                and not Task.objects.filter(
                    workflow_id=instance.workflow_id, run=instance.run
                )
                .exclude(state__in=FINAL_STATES)
                .exists()
            ):
                instance.workflow.finished = timezone.now()
                instance.workflow.save()

        # We already checked earlier that the task was in a final state.
        # If this state is both final and not completed, then we should trigger the task failure signal.
        if instance.state != State.Completed:
            task_failure.send_robust(self.__class__, task=instance)

        return instance


class TaskTinySerializer(TaskSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance with only its state.
    Used by humans to update a task.
    """

    state = EnumField(State)

    class Meta:
        model = Task
        fields = ("id", "state")
        read_only_fields = ("id",)

    def update(self, instance: Task, validated_data) -> Task:
        new_state = validated_data.get('state')
        if new_state == State.Stopping and instance.state != State.Running:
            raise ValidationError("You can only stop a 'Running' task")

        if new_state == State.Pending:
            if instance.state not in FINAL_STATES:
                raise ValidationError(
                    "You can only restart a task with a state equal to 'Completed', 'Failed', 'Error' or 'Stopped'."
                )
            instance.agent = None

        if new_state in FINAL_STATES and new_state != State.Completed:
            task_failure.send_robust(self.__class__, task=instance)

        return super().update(instance, validated_data)


class AgentStateSerializer(AgentLightSerializer):
    """
    Serialize an :class:`~arkindex.ponos.models.Agent` with its state information
    And the GPU state
    """

    running_tasks_count = serializers.IntegerField(min_value=0)

    class Meta(AgentLightSerializer.Meta):
        fields = AgentLightSerializer.Meta.fields + ("active", "running_tasks_count")


class AgentDetailsSerializer(AgentLightSerializer):
    """
    Serialize an :class:`~arkindex.ponos.models.Agent` with its running tasks
    """

    running_tasks = serializers.SerializerMethodField()

    class Meta(AgentLightSerializer.Meta):
        fields = AgentLightSerializer.Meta.fields + ("active", "running_tasks")

    @extend_schema_field(TaskLightSerializer(many=True))
    def get_running_tasks(self, agent):
        running_tasks = agent.tasks.filter(state=State.Running)
        task_serializer = TaskLightSerializer(
            instance=running_tasks, many=True, context=self.context
        )
        return task_serializer.data


class GPUCreateSerializer(serializers.Serializer):
    """
    Serialize a :class:`~arkindex.ponos.models.GPU` attached to an Agent
    """

    id = serializers.UUIDField()
    name = serializers.CharField()
    index = serializers.IntegerField()
    ram_total = serializers.IntegerField()


class AgentCreateSerializer(serializers.ModelSerializer):
    """
    Serializer used to register a new :class:`~arkindex.ponos.models.Agent`.
    """

    public_key = PublicKeyField(write_only=True)
    derivation = Base64Field(write_only=True)
    access_token = serializers.CharField(read_only=True, source="token.access_token")
    refresh_token = serializers.CharField(read_only=True, source="token")
    gpus = GPUCreateSerializer(many=True)

    class Meta:
        model = Agent
        fields = (
            "id",
            "farm",
            "access_token",
            "refresh_token",
            "public_key",
            "derivation",
            "hostname",
            "cpu_cores",
            "cpu_frequency",
            "ram_total",
            "cpu_load",
            "ram_load",
            "last_ping",
            "gpus",
        )
        read_only_fields = (
            "id",
            "access_token",
            "refresh_token",
            "last_ping",
        )

    def create(self, validated_data):
        """Create the required agent and its GPUs"""
        gpus = validated_data.pop("gpus", None)

        with transaction.atomic(using="default"):

            # Create Agent as usual
            agent = super().create(validated_data)

            # Create or update GPUs
            # When an agent's private key is regenerated on the same host, an existing GPU UUID might be sent,
            # so we need to handle existing GPUs to avoid errors.
            for gpu_data in gpus:
                GPU.objects.update_or_create(
                    id=gpu_data.pop("id"), defaults=dict(agent=agent, **gpu_data)
                )

        return agent

    def update(self, instance, validated_data):
        """Create the required agent and its GPUs"""
        gpus = validated_data.pop("gpus", None)

        with transaction.atomic(using="default"):

            # Update Agent as usual
            agent = super().update(instance, validated_data)

            # Delete existing GPUs
            agent.gpus.all().delete()

            # Create Gpus - the ID should not evolve as it's provided
            # by the host
            for gpu_data in gpus:
                agent.gpus.create(**gpu_data)

        return agent

    def validate(self, data):
        if not check_agent_key(
            data["public_key"], data["derivation"], data["farm"].seed
        ):
            raise serializers.ValidationError("Key verification failed")
        del data["derivation"]

        # Turn the public key back into a string to be saved in the agent
        # TODO: Use a custom Django model field for this?
        data["public_key"] = PublicKeyField().to_representation(data["public_key"])

        # Generate the agent ID as a MD5 hash of its public key
        data["id"] = uuid.UUID(
            hashlib.md5(data["public_key"].encode("utf-8")).hexdigest()
        )

        # Set last ping as agent is about to be registered
        data["last_ping"] = timezone.now()

        return data


class WorkflowSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Workflow` instance with its tasks.
    """

    tasks = TaskLightSerializer(many=True, read_only=True)
    farm = FarmSerializer(read_only=True)
    state = EnumField(State)

    class Meta:
        model = Workflow
        fields = (
            "id",
            "created",
            "finished",
            "state",
            "farm",
            "tasks",
        )
        read_only_fields = (
            "id",
            "created",
            "finished",
            "farm",
            "tasks",
        )

    def validate_state(self, state: State) -> None:
        """
        When performing updates, prevents updating to any state other than :attr:`~arkindex.ponos.models.State.Stopping`
        and restricts to :attr:`~arkindex.ponos.models.State.Pending` or :attr:`~arkindex.ponos.models.State.Running` workflows.
        """
        if state != State.Stopping:
            raise ValidationError("Can only change the state to 'stopping'")
        if self.instance.state not in (State.Pending, State.Running):
            raise ValidationError(
                "Cannot stop a {} workflow".format(self.instance.state.value)
            )


class ActionSerializer(serializers.Serializer):
    """
    Serializes an :const:`~arkindex.ponos.models.Action` instance.
    """

    action = EnumField(ActionType)
    task_id = serializers.SerializerMethodField()

    @extend_schema_field(serializers.UUIDField(allow_null=True))
    def get_task_id(self, obj):
        if not obj.task:
            return
        return str(obj.task.id)


class AgentActionsSerializer(serializers.Serializer):
    """
    Serializes multiple next actions for an agent.
    """

    actions = ActionSerializer(many=True, source="next_actions")


class TaskDefinitionSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance
    to provide startup information to assigned agents.
    """

    env = serializers.DictField(default={})
    workflow_id = serializers.UUIDField()
    agent_id = serializers.PrimaryKeyRelatedField(queryset=Agent.objects.all())
    image_artifact_url = serializers.SerializerMethodField()
    s3_logs_put_url = serializers.SerializerMethodField()
    extra_files = serializers.DictField(default={})

    @extend_schema_field(serializers.URLField(allow_null=True))
    def get_image_artifact_url(self, task):
        """Build url on the API to get a fresh download link"""
        if not task.image_artifact:
            return
        return self.context["request"].build_absolute_uri(
            reverse(
                "api:task-artifact-download",
                kwargs={
                    "pk": task.image_artifact.task_id,
                    "path": task.image_artifact.path,
                },
            )
        )

    @extend_schema_field(serializers.URLField(allow_null=True))
    def get_s3_logs_put_url(self, obj):
        if "request" not in self.context or self.context["request"].user != obj.agent:
            return
        return obj.s3_logs_put_url

    class Meta:
        model = Task
        fields = (
            "id",
            "slug",
            "image",
            "command",
            "env",
            "shm_size",
            "has_docker_socket",
            "image_artifact_url",
            "agent_id",
            "s3_logs_put_url",
            "parents",
            "workflow_id",
            "gpu_id",
            "extra_files",
        )
        read_only_fields = fields


class ArtifactSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Artifact` instance to allow
    agents to create them, and users to list and retrieve them
    """

    s3_put_url = serializers.SerializerMethodField()
    url = serializers.SerializerMethodField()

    class Meta:
        model = Artifact
        fields = (
            "id",
            "path",
            "size",
            "content_type",
            "url",
            "s3_put_url",
            "created",
            "updated",
        )
        read_only_fields = (
            "id",
            "s3_put_url",
            "created",
            "updated",
        )

    def get_url(self, obj) -> str:
        """Build url on the API to get a fresh download link"""
        return self.context["request"].build_absolute_uri(
            reverse(
                "api:task-artifact-download",
                kwargs={"pk": obj.task_id, "path": obj.path},
            )
        )

    @extend_schema_field(serializers.URLField(allow_null=True))
    def get_s3_put_url(self, obj):
        """Only add the PUT url to store the file for agents during creation"""
        request = self.context.get("request")
        if not request or request.method != "POST":
            return
        return obj.s3_put_url

    def validate_path(self, path):
        """Check that no artifacts with this path already exist in DB"""
        task = self.context["view"].get_task()
        if task.artifacts.filter(path=path).exists():
            raise ValidationError("An artifact with this path already exists")

        return path


class NewTaskSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Task` instance to permit creation by a parent.
    """

    workflow_id = serializers.UUIDField()
    command = serializers.CharField(required=False)
    env = serializers.DictField(
        child=serializers.CharField(), required=False, default={}
    )
    has_docker_socket = serializers.BooleanField(default=False)

    class Meta:
        model = Task
        fields = (
            "id",
            "workflow_id",
            "slug",
            "parents",
            "image",
            "command",
            "env",
            "run",
            "depth",
            "has_docker_socket",
        )
        read_only_fields = (
            "id",
            "run",
            "depth",
        )

    def validate(self, data):
        parents = data["parents"]

        ids = {parent.workflow.id for parent in parents}
        if len(ids) != 1 or str(ids.pop()) != str(data["workflow_id"]):
            raise ValidationError(
                "All parents must be in the same workflow as the child task"
            )

        runs = {parent.run for parent in parents}
        if len(runs) != 1:
            raise ValidationError(
                "All parents must have the same run in the given workflow"
            )
        data["run"] = runs.pop()

        if Task.objects.filter(
            workflow_id=data["workflow_id"], run=data["run"], slug=data["slug"]
        ).exists():
            raise ValidationError(
                f"A task with the `{data['slug']}` slug already exists in run {data['run']}."
            )

        data["depth"] = max(parent.depth for parent in parents) + 1

        # Build task environment from PONOS_DEFAULT_ENV if the env is not (fully) defined in data
        data["env"] = {**settings.PONOS_DEFAULT_ENV, **data["env"]}
        # Get process and corpus id from any of the parent tasks since it's the same for all of them
        data["env"]["ARKINDEX_PROCESS_ID"] = parents[0].env['ARKINDEX_PROCESS_ID']
        data["env"]["ARKINDEX_CORPUS_ID"] = parents[0].env.get('ARKINDEX_CORPUS_ID', None)

        # Set the task token manually so that we can immediately copy it to the environment variables,
        # just like what is done in Workflow.build_tasks()
        data["token"] = task_token_default()
        data["env"]['ARKINDEX_TASK_TOKEN'] = data['token']

        return super().validate(data)


class ClearTextSecretSerializer(serializers.ModelSerializer):
    """
    Serializes a :class:`~arkindex.ponos.models.Secret` instance with its content in cleartext.
    """

    content = serializers.SerializerMethodField(read_only=True)

    def get_content(self, secret) -> str:
        return secret.decrypt()

    class Meta:
        model = Secret
        fields = ("id", "name", "content")
        read_only_fields = ("id", "content")
