import uuid
from datetime import timedelta
from unittest.mock import patch

from django.test import TestCase
from django.utils import timezone

from arkindex.ponos.models import ACTIVE_STATES, FINAL_STATES, Agent, Farm, State, Task, Workflow
from arkindex.ponos.models import timezone as model_tz


class TasksAttributionTestCase(TestCase):
    """
    Ponos server distribute tasks equally among agents.
    """

    @classmethod
    def setUpTestData(cls):
        cls.farm = Farm.objects.create(name="testfarm")
        cls.workflow = Workflow.objects.create(farm=cls.farm)

    def _run_tasks(self, tasks):
        """
        Mark a list of tasks as running
        """
        for t in tasks:
            t.state = State.Running
            t.save()

    def _build_agent(self, **kwargs):
        """
        Creates an agent
        Default values may be overridden if passed as kwargs
        """
        params = {
            "hostname": "test_host",
            "cpu_cores": 2,
            "cpu_frequency": 4.2e9,
            "public_key": "",
            "farm": self.farm,
            "ram_total": 2e9,
            "last_ping": timezone.now(),
            "ram_load": 0.49,
            "cpu_load": 0.99,
        }
        params.update(kwargs)
        return Agent.objects.create(**params)

    def _add_pending_tasks(self, number, slug_ext="", **kwargs):
        """
        Creates pending tasks on the system
        """
        params = {
            "run": 0,
            "depth": 0,
            "workflow": self.workflow,
            "state": State.Pending,
        }
        params.update(kwargs)
        return Task.objects.bulk_create(
            Task(**params, slug=f"slug{slug_ext}_{i}") for i in range(1, number + 1)
        )

    def _active_agent(self, agent):
        from arkindex.ponos.models import AGENT_TIMEOUT

        return timezone.now() - agent.last_ping < AGENT_TIMEOUT

    def test_distribute_tasks(self):
        """
        Multiple tasks are attributed to 3 registered agents with the same capacity
        """
        agent_1, agent_2, agent_3 = [
            # Build three agents with 10 free CPU cores and 10Go of available RAM
            self._build_agent(
                hostname=f"agent_{i}",
                cpu_cores=11,
                ram_total=20e9,
            )
            for i in range(1, 4)
        ]
        self.assertEqual(
            [a.hostname for a in Agent.objects.all() if self._active_agent(a) is True],
            ["agent_1", "agent_2", "agent_3"],
        )
        # Add 9 pending tasks to the system
        self._add_pending_tasks(9)

        # As agents are registered, they should retrieve 3 tasks each
        tasks = agent_1.next_tasks()
        self.assertEqual(len(tasks), 3)
        self._run_tasks(tasks)

        # Agent 1 claims tasks before agent 2 but has a too high load to retrieve any task
        agent_1.ram_load = 0.75
        agent_1.cpu_load = 3.99
        agent_1.save()
        tasks = agent_1.next_tasks()
        self.assertEqual(len(tasks), 0)
        bg_tasks = agent_2.next_tasks()
        self.assertEqual(len(bg_tasks), 3)

        # Agent 1 claims tasks before agent 3
        tasks = agent_1.next_tasks()
        self.assertEqual(len(tasks), 0)
        tasks = agent_3.next_tasks()
        self.assertEqual(len(tasks), 3)

        # No agent updated its state before agent_2 retrieved its tasks
        self.assertEqual(len(set(t.id for t in [*bg_tasks, *tasks])), 6)

        self._run_tasks(bg_tasks)
        self._run_tasks(tasks)
        self.assertEqual(Task.objects.filter(state=State.Pending).count(), 0)

    def test_distribute_tasks_asymetric(self):
        """
        Use case when two agents with different capacity are attributed
        tasks equivalent to a third of the system capacity
        """
        # Agent 1 has 10 free "slots"
        agent_1 = self._build_agent(hostname="agent_1", cpu_cores=11, ram_total=20e9)
        # Agent 2 has only 3 free "slots"
        agent_2 = self._build_agent(hostname="agent_2", cpu_cores=4, ram_total=6e9)
        # Add 5 pending tasks to the system
        self._add_pending_tasks(5)

        # The best strategy is to attribute 4 tasks to the first agent and 1 to the second
        tasks = agent_1.next_tasks()
        self.assertEqual(len(tasks), 4)
        self._run_tasks(tasks)

        agent_1.ram_load = 0.7
        agent_1.cpu_load = 3.99
        agent_1.save()
        tasks = agent_1.next_tasks()
        self.assertEqual(len(tasks), 0)
        tasks = agent_2.next_tasks()
        self.assertEqual(len(tasks), 1)
        self._run_tasks(tasks)
        self.assertEqual(Task.objects.filter(state=State.Pending).count(), 0)

    def test_distribute_tasks_determinist(self):
        """
        Tasks should be distributed in a determinist way depending
        on the overall load and pending tasks queue
        Tests that an agent polling task twice retrieves the same tasks
        """
        _, _, agent = [
            self._build_agent(
                hostname=f"agent_{i}",
                cpu_cores=11,
                ram_total=20e9,
            )
            for i in range(1, 4)
        ]
        self._add_pending_tasks(9)
        self.assertEqual(
            [a.hostname for a in Agent.objects.all() if self._active_agent(a) is True],
            ["agent_1", "agent_2", "agent_3"],
        )

        tasks = agent.next_tasks()
        self.assertEqual(len(tasks), 3)

        new_tasks = agent.next_tasks()
        self.assertCountEqual(tasks, new_tasks)

    @patch("arkindex.ponos.models.AGENT_TIMEOUT", timedelta(seconds=1))
    def test_non_active_agent(self):
        """
        If an agent does not respond, task should be attributed to other running agents
        """
        agent_1, agent_2, agent_3 = [
            self._build_agent(
                hostname=f"agent_{i}",
                cpu_cores=11,
                ram_total=20e9,
            )
            for i in range(1, 4)
        ]
        self.assertEqual(
            [a.hostname for a in Agent.objects.all() if self._active_agent(a) is True],
            ["agent_1", "agent_2", "agent_3"],
        )
        # Add 9 pending tasks to the system
        self._add_pending_tasks(9)

        # As agents are registered, they should retrieve 3 tasks each
        tasks_agent_1 = agent_1.next_tasks()
        self.assertEqual(len(tasks_agent_1), 3)

        tasks = agent_2.next_tasks()
        self.assertEqual(len(tasks), 3)

        self._run_tasks([*tasks_agent_1, *tasks])

        # Jump in the future. Agent 3 did never replied
        future_now = timezone.now() + timedelta(seconds=2)
        Agent.objects.filter(hostname__in=["agent_1", "agent_2"]).update(
            ram_load=0.64,
            cpu_load=3.99,
            last_ping=future_now,
        )

        with patch.object(model_tz, "now") as now_mock:
            now_mock.return_value = future_now
            self.assertEqual(
                [
                    a.hostname
                    for a in Agent.objects.all()
                    if self._active_agent(a) is True
                ],
                ["agent_1", "agent_2"],
            )

            # 3 tasks should be distributed among 2 agents with similar load
            tasks_agent_1 = agent_1.next_tasks()
            self.assertEqual(len(tasks_agent_1), 2)

            tasks = agent_2.next_tasks()
            self.assertEqual(len(tasks), 1)

            self._run_tasks([*tasks_agent_1, *tasks])

        self.assertEqual(Task.objects.filter(state=State.Pending).count(), 0)

    def test_filter_active_agents(self):
        """
        Assert that the DB cost to attribute tasks does
        not increase with tasks or agents number
        """
        agent = self._build_agent()
        tasks = self._add_pending_tasks(20)

        with self.assertNumQueries(4):
            tasks = agent.next_tasks()
            self.assertEqual(len(tasks), 1)

        # Build some inefficient agents
        for i in range(1, 20):
            self._build_agent(
                hostname=f"agent_{i}",
                cpu_cores=2,
                ram_total=2e9,
            )

        self.assertEqual(Agent.objects.count(), 20)
        with self.assertNumQueries(4):
            tasks = agent.next_tasks()
            self.assertEqual(len(tasks), 1)

    def test_gpu_assignation(self):
        """
        Check a GPU enabled agent gets a GPU task
        """
        # Agent normal has 20 free "slots"
        # It should get all the tasks
        agent_normal = self._build_agent(
            hostname="agent_normal", cpu_cores=20, ram_total=20e9
        )

        # But this one has a GPU: it will get GPU tasks in priority
        agent_gpu = self._build_agent(hostname="agent_2", cpu_cores=10, ram_total=6e9)
        agent_gpu.gpus.create(id=uuid.uuid4(), index=0, ram_total=8e9, name="Fake GPU")

        # Add 6 normal + 1 GPU pending tasks to the system
        tasks = self._add_pending_tasks(7)
        task_gpu = tasks[-1]

        # Require gpu on that task
        task_gpu.command = "/usr/bin/nvidia-smi"
        task_gpu.requires_gpu = True
        task_gpu.save()

        # Normal agent should eat up most of the normal tasks
        tasks = agent_normal.next_tasks()
        self.assertEqual(len(tasks), 5)
        self.assertFalse(any(t.requires_gpu for t in tasks))
        self.assertFalse(any(t.gpu for t in tasks))
        self._run_tasks(tasks)

        # GPU agent should then get the GPU task
        tasks = agent_gpu.next_tasks()
        self.assertEqual(len(tasks), 1)
        self.assertTrue(all(t.requires_gpu for t in tasks))
        self.assertTrue(all(t.gpu for t in tasks))
        self._run_tasks(tasks)

        # Consume last normal task
        tasks = agent_normal.next_tasks()
        self.assertEqual(len(tasks), 1)
        self._run_tasks(tasks)

        # No more tasks left
        self.assertEqual(Task.objects.filter(state=State.Pending).count(), 0)

    def test_gpu_with_final_tasks_available(self):
        """
        A GPU that is assigned to some final tasks should still be available
        """
        gpu_agent = self._build_agent(hostname='matrix', cpu_cores=99, ram_total=9e12)
        gpu = gpu_agent.gpus.create(id=uuid.uuid4(), index=0, ram_total=16e12, name='My cool GPU')

        # Create some final tasks that use this GPU
        Task.objects.bulk_create(
            Task(
                depth=0,
                run=0,
                workflow=self.workflow,
                requires_gpu=True,
                slug=f'task{i}',
                state=state,
                gpu=gpu,
            )
            for i, state in enumerate(FINAL_STATES)
        )

        # Create another task that needs a GPU and should be assigned to this agent
        new_task = self.workflow.tasks.create(
            run=0,
            depth=0,
            slug='new_task',
            requires_gpu=True,
            state=State.Unscheduled,
        )

        self.assertListEqual(gpu_agent.next_tasks(), [new_task])

        new_task.refresh_from_db()
        # The agent and GPU are assigned in Agent.next_actions, not next_tasks
        self.assertIsNone(new_task.agent)
        self.assertIsNone(new_task.gpu)

    def test_gpu_with_active_task_unavailable(self):
        """
        A GPU that is assigned to any non-final task should not be available
        """
        gpu_agent = self._build_agent(hostname='matrix', cpu_cores=99, ram_total=9e12)
        gpu = gpu_agent.gpus.create(id=uuid.uuid4(), index=0, ram_total=16e12, name='My cool GPU')

        for state in ACTIVE_STATES:
            self.workflow.tasks.all().delete()
            with self.subTest(state=state):
                # Create a task that is running on this agent's GPU
                self.workflow.tasks.create(
                    run=0,
                    depth=0,
                    slug='current_task',
                    requires_gpu=True,
                    agent=gpu_agent,
                    gpu=gpu,
                    state=state,
                )

                # Create another task that needs a GPU
                new_task = self.workflow.tasks.create(
                    run=0,
                    depth=0,
                    slug='new_task',
                    requires_gpu=True,
                    state=State.Unscheduled,
                )

                self.assertListEqual(gpu_agent.next_tasks(), [])

                new_task.refresh_from_db()
                self.assertIsNone(new_task.agent)
                self.assertIsNone(new_task.gpu)

    def test_multiple_farm_task_assignation(self):
        """
        Distribute tasks depending on farm
        """
        # Create one big agent on the test farm
        test_agent = self._build_agent(cpu_cores=20, ram_total=64e9)
        self._add_pending_tasks(3)
        # and 3 small agents on the corn farm with a capacity for one task each
        corn_farm = Farm.objects.create(name="Corn farm")
        corn_agent_1, corn_agent_2 = [
            self._build_agent(
                hostname=f"agent_{i}",
                cpu_cores=2,
                ram_total=2e9,
                farm=corn_farm,
            )
            for i in range(1, 3)
        ]

        corn_workflow = corn_farm.workflows.create()
        tasks = self._add_pending_tasks(3, workflow=corn_workflow)

        self.assertEqual(Task.objects.count(), 6)
        self.assertEqual(
            len([agent for agent in Agent.objects.all() if self._active_agent(agent)]),
            3,
        )

        # All except one task in the corn farm will be distributed
        tasks = test_agent.next_tasks()
        self.assertEqual(len(tasks), 3)
        self.assertEqual(
            set([task.workflow.farm_id for task in tasks]), set([self.farm.id])
        )

        corn_tasks_1 = corn_agent_1.next_tasks()
        self.assertEqual(len(corn_tasks_1), 1)
        corn_tasks_2 = corn_agent_2.next_tasks()
        self.assertEqual(len(corn_tasks_2), 1)

        self._run_tasks([*tasks, *corn_tasks_1, *corn_tasks_2])

        # Update corn agents loads
        Agent.objects.filter(farm=corn_farm).update(ram_load=0.95, cpu_load=1.9)

        # No agent can retrieve the last pending task
        self.assertEqual(Task.objects.filter(state=State.Pending).count(), 1)
        for agent in Agent.objects.all():
            self.assertEqual(len(agent.next_tasks()), 0)

    def test_next_tasks_ordering(self):
        agent_1 = self._build_agent(hostname="agent_1", cpu_cores=11, ram_total=20e9)
        self.assertEqual(
            [a.hostname for a in Agent.objects.all() if self._active_agent(a) is True],
            ["agent_1"],
        )
        # Add 3 pending low priority tasks to the system
        lp_tasks = self._add_pending_tasks(3, slug_ext="_lp", priority="1")
        # Add 3 pending normal priority tasks to the system
        np_tasks = self._add_pending_tasks(3, slug_ext="_np")
        # Add 3 pending high priority tasks to the system
        hp_tasks = self._add_pending_tasks(3, slug_ext="_hp", priority="100")

        # As agents are registered, they should retrieve 3 tasks each
        tasks = agent_1.next_tasks()
        self.assertEqual(len(tasks), 9)
        # First three tasks should be the high priority ones
        self.assertEqual(tasks[:3], hp_tasks)
        # Second three tasks should be the normal priority ones
        self.assertEqual(tasks[3:6], np_tasks)
        # Last three tasks should be the low priority ones
        self.assertEqual(tasks[6:], lp_tasks)
