from unittest.mock import patch

from django.test import TestCase
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from arkindex.ponos.models import Agent, Farm, State, Workflow


class WorkflowTestCase(TestCase):
    """
    Create some workflows & tasks
    """

    @classmethod
    def setUpTestData(cls):
        cls.farm = Farm.objects.create(name="testfarm")
        # Create a fake Agent instance
        cls.agent = Agent.objects.create(
            hostname="test_agent",
            cpu_cores=2,
            cpu_frequency=1e9,
            public_key="",
            farm=cls.farm,
            ram_total=2e9,
            last_ping=timezone.now(),
        )

    def test_workflow_running_override(self):
        """
        Test that a single running task in a workflow will override any other state
        """
        w = Workflow.objects.create(farm=self.farm)
        t1 = w.tasks.create(image='hello-world', run=0, depth=0, slug='t1')
        t2 = w.tasks.create(image='hello-world', run=0, depth=1, slug='t2')
        t3 = w.tasks.create(image='hello-world', run=0, depth=2, slug='t3')
        t2.parents.add(t1)
        t3.parents.add(t2)

        self.assertEqual(w.state, State.Unscheduled)

        t1.state = State.Running
        t1.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Error
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Failed
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Stopping
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Stopped
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Pending
        t2.save()
        self.assertEqual(w.state, State.Running)

        t2.state = State.Completed
        t3.state = State.Completed
        t2.save()
        t3.save()
        self.assertEqual(w.state, State.Running)

    @patch("arkindex.ponos.models.s3")
    @patch("arkindex.ponos.models.Task.s3_logs_get_url")
    def test_task_parents_update(self, s3_mock, s3_logs_mock):
        w = Workflow.objects.create(farm=self.farm)

        parent1 = w.tasks.create(slug="parent1", image='hello-world', run=0, depth=0)
        parent2 = w.tasks.create(slug="parent2", image='hello-world', run=0, depth=0)
        child = w.tasks.create(slug="test", image='hello-world', run=0, depth=1)
        child.parents.set([parent1, parent2])

        response = self.client.get(
            reverse("api:agent-actions"),
            HTTP_AUTHORIZATION="Bearer {}".format(self.agent.token.access_token),
            data={"cpu_load": 0.99, "ram_load": 0.49},
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        actions = response.json()["actions"]
        self.assertEqual(len(actions), 1)
        self.assertNotEqual(actions[0]["task_id"], str(child.id))

        parent1.state = State.Completed
        parent1.save()

        parent2.agent = self.agent
        parent2.save()

        self.client.get(
            reverse("api:agent-actions"),
            HTTP_AUTHORIZATION="Bearer {}".format(self.agent.token.access_token),
        )
        # Completing 2nd parent should change test_task state to pending
        response = self.client.patch(
            reverse("api:task-details", kwargs={"pk": str(parent2.id)}),
            HTTP_AUTHORIZATION="Bearer {}".format(self.agent.token.access_token),
            data={"state": State.Completed.value},
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        child.refresh_from_db()
        self.assertEqual(child.state, State.Pending)
        self.assertEqual(w.state, State.Pending)

    def test_workflow_retry_updates_finished(self):
        w = Workflow.objects.create(farm=self.farm)
        w.tasks.create(image='hello-world', slug='task', state=State.Completed, run=0, depth=0)
        w.finished = timezone.now()
        w.save()

        with self.assertNumQueries(7):
            w.retry()
        self.assertEqual(w.tasks.count(), 2)
        self.assertIsNone(w.finished)

    def test_workflow_stop_unscheduled_updates_finished(self):
        w = Workflow.objects.create(farm=self.farm)
        task = w.tasks.create(image='hello-world', slug='task', state=State.Unscheduled, run=0, depth=0)

        # Stopping the workflow causes all tasks to immediately be Stopped
        w.stop()
        self.assertIsNotNone(w.finished)
        task.refresh_from_db()
        self.assertEqual(task.state, State.Stopped)
