import logging

from django.conf import settings
from django.db import connections, models
from django.utils.functional import cached_property

logger = logging.getLogger(__name__)


class ActivityManager(models.Manager):
    """Model management for worker activities"""

    def bulk_insert(self, worker_version_id, process_id, configuration_id, elements_qs):
        """
        Create initial worker activities from a queryset of elements in a efficient way.
        Due to the possible large amount of elements, we use a bulk insert from the elements query (best performances).
        The `ON CONFLICT` clause allows either skipping existing activities if they are in a `Started` or `Processed`
        state, or stealing them from another process if they are `Queued` or `Error`.
        """
        sql, params = elements_qs.distinct().values('id').query.sql_with_params()
        select_params = (worker_version_id, configuration_id, process_id) + params

        # With ON CONFLICT, the target constraint is only optional when the action is DO NOTHING.
        # For DO UPDATE, the target can either be a constraint by name with `ON CONSTRAINT <name>`,
        # or the definition of an index that PostgreSQL will look for, using a list of columns,
        # a collation, an operator class, a predicate, etc.
        # `ON CONSTRAINT <name>` does not work here because it is unable to detect partial constraints.
        # We therefore have to explicitly define one of the two partial unique constraints ourselves.
        if configuration_id is None:
            conflict_target = '(worker_version_id, element_id) WHERE configuration_id IS NULL'
        else:
            conflict_target = '(worker_version_id, element_id, configuration_id) WHERE configuration_id IS NOT NULL'

        with connections['default'].cursor() as cursor:
            cursor.execute(
                f"""
                INSERT INTO process_workeractivity
                    (element_id, worker_version_id, configuration_id, state, process_id, id, created, updated)
                SELECT
                    elt.id, %s, %s, 'queued', %s, uuid_generate_v4(), current_timestamp, current_timestamp
                FROM ({sql}) AS elt
                ON CONFLICT {conflict_target} DO UPDATE SET
                    process_id = EXCLUDED.process_id,
                    state = EXCLUDED.state,
                    created = EXCLUDED.created
                WHERE process_workeractivity.state <> 'started' AND process_workeractivity.state <> 'processed'
                """,
                select_params)


class CorpusWorkerVersionManager(models.Manager):

    def rebuild(self):
        """
        Rebuild the corpus worker versions cache from all ML results.
        """
        from arkindex.documents.models import (
            Classification,
            Element,
            Entity,
            MetaData,
            Transcription,
            TranscriptionEntity,
        )

        querysets = [
            Element.objects.exclude(worker_version_id=None).values_list('corpus_id', 'worker_version_id'),
            Transcription.objects.exclude(worker_version_id=None).values_list('element__corpus_id', 'worker_version_id'),
            Entity.objects.exclude(worker_version_id=None).values_list('corpus_id', 'worker_version_id'),
            TranscriptionEntity.objects.exclude(worker_version_id=None).values_list('entity__corpus_id', 'worker_version_id'),
            Classification.objects.exclude(worker_version_id=None).values_list('element__corpus_id', 'worker_version_id'),
            MetaData.objects.exclude(worker_version_id=None).values_list('element__corpus_id', 'worker_version_id'),
        ]

        for i, queryset in enumerate(querysets, start=1):
            logger.info(f'Rebuilding cache from {queryset.model.__name__} ({i}/{len(querysets)})')
            self.bulk_create([
                self.model(corpus_id=corpus_id, worker_version_id=worker_version_id)
                for corpus_id, worker_version_id in queryset.distinct()
            ], ignore_conflicts=True)


class WorkerVersionManager(models.Manager):

    @cached_property
    def imports_version(self):
        """
        WorkerVersion used for all import tasks.
        """
        return (
            self
            # Required by WorkerRun.build_summary
            .select_related('worker', 'revision')
            .prefetch_related('revision__refs')
            .get(id=settings.IMPORTS_WORKER_VERSION)
        )

    @cached_property
    def transkribus_version(self):
        """
        WorkerVersion used for all Transkribus imports.
        """
        return (
            self
            # Required by WorkerRun.build_summary
            .select_related('worker', 'revision')
            .prefetch_related('revision__refs')
            .get(id=settings.TRANSKRIBUS_WORKER_VERSION)
        )
