from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers

from arkindex.documents.models import Corpus
from arkindex.ponos.utils import is_admin_or_ponos_task
from arkindex.process.models import DataFile
from arkindex.project.aws import S3FileStatus
from arkindex.project.serializer_fields import EnumField


class DataFileSerializer(serializers.ModelSerializer):
    """
    Serialize a single uploaded file
    """

    status = EnumField(S3FileStatus)
    s3_url = serializers.SerializerMethodField()

    class Meta:
        model = DataFile
        fields = (
            'id',
            'name',
            'content_type',
            'size',
            'status',
            's3_url',
        )
        read_only_fields = ('id', 'size', 'content_type', 's3_url', )

    def validate_status(self, value):
        if value == S3FileStatus.Checked:
            # Status has been resquested to be checked, perform validation
            try:
                self.instance.perform_check(raise_exc=True)
            except (AssertionError, ValueError) as e:
                raise serializers.ValidationError(str(e))
        return value

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_url(self, obj):
        if 'request' not in self.context:
            return
        # Only allow the S3 URL for internal users or admins
        if is_admin_or_ponos_task(self.context['request']):
            return obj.s3_url


class DataFileCreateSerializer(serializers.ModelSerializer):
    """
    Serialize a Datafile creation with Amazon S3 PUT uri
    """

    status = EnumField(S3FileStatus, read_only=True)
    s3_put_url = serializers.SerializerMethodField()
    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())

    class Meta:
        model = DataFile
        fields = (
            'id',
            'name',
            'content_type',
            'size',
            'corpus',
            'status',
            's3_url',
            's3_put_url',
        )
        read_only_fields = ('id', 'status', 's3_url', 's3_put_url')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        self.fields['corpus'].queryset = Corpus.objects.writable(self.context['request'].user)

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_put_url(self, obj):
        if obj.status == S3FileStatus.Checked:
            return None
        return obj.s3_put_url
