import uuid
from datetime import datetime, timezone
from unittest.mock import MagicMock, call, patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.process.models import (
    ActivityState,
    Process,
    ProcessMode,
    Repository,
    Revision,
    WorkerActivity,
    WorkerActivityState,
    WorkerConfiguration,
    WorkerType,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.process.tasks import initialize_activity
from arkindex.project.tests import FixtureTestCase


class TestWorkerActivity(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.element = Element.objects.get(name='Volume 1, page 2r')
        cls.process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=cls.user,
            corpus=cls.corpus,
        )
        cls.configuration = WorkerConfiguration.objects.create(worker=cls.worker_version.worker, name='A config', configuration={'a': 'b'})
        cls.process.worker_runs.create(version=cls.worker_version, parents=[], configuration=cls.configuration)
        cls.worker_type = WorkerType.objects.get(slug='recognizer')
        cls.process.start()
        cls.task = cls.process.workflow.tasks.get(slug=cls.worker_version.slug)

    def setUp(self):
        super().setUp()
        # Create a queued activity for this element
        self.activity = self.element.activities.create(
            process=self.process,
            worker_version=self.worker_version,
            state=WorkerActivityState.Queued,
            configuration=self.configuration
        )

    def test_bulk_insert_activity_no_configuration(self):
        """
        Bulk insert worker activities for acts
        """
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Queued, process=self.process).count(), 1)
        elements_qs = Element.objects.filter(type__slug='act', type__corpus_id=self.corpus.id)
        params = {
            'worker_version_id': self.worker_version.id,
            'corpus_id': self.corpus.id,
            'process_id': self.process.id,
        }
        with self.assertExactQueries('workeractivity_bulk_insert_no_configuration.sql', params=params):
            WorkerActivity.objects.bulk_insert(self.worker_version.id, self.process.id, None, elements_qs)
        self.assertEqual(elements_qs.count(), 5)
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Queued, process=self.process).count(), 6)

    def test_bulk_insert_activity_duplicate_elements(self):
        """
        WorkerActivity.bulk_insert should exclude duplicate elements
        """
        element_type = self.corpus.types.first()
        parent1 = self.corpus.elements.create(type=element_type, name='Parent 1')
        parent2 = self.corpus.elements.create(type=element_type, name='Parent 2')
        element = self.corpus.elements.create(type=element_type, name='Element')
        child = self.corpus.elements.create(type=element_type, name='Child')
        element.add_parent(parent1)
        element.add_parent(parent2)
        child.add_parent(element)
        elements_qs = Element.objects.filter(paths__path__contains=[element.id], name='Child')
        # `child` has two paths that both contain the ID of `element`, because `element` has two parents,
        # so filtering on paths__path will duplicate the child
        self.assertEqual(elements_qs.count(), 2)
        self.activity.delete()

        WorkerActivity.objects.bulk_insert(
            self.worker_version.id,
            self.process.id,
            self.configuration.id,
            elements_qs,
        )
        self.assertEqual(self.process.activities.get().element_id, child.id)

    def test_bulk_insert_activity_children(self):
        """
        Bulk insert worker activities for acts
        """
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Queued, process=self.process, configuration=self.configuration).count(), 1)
        elements_qs = Element.objects.filter(type__slug='act', type__corpus_id=self.corpus.id)
        params = {
            'worker_version_id': self.worker_version.id,
            'configuration_id': self.configuration.id,
            'corpus_id': self.corpus.id,
            'process_id': self.process.id,
        }
        with self.assertExactQueries('workeractivity_bulk_insert.sql', params=params):
            WorkerActivity.objects.bulk_insert(self.worker_version.id, self.process.id, self.configuration.id, elements_qs)
        self.assertEqual(elements_qs.count(), 5)
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Queued, process=self.process, configuration=self.configuration).count(), 6)

    def test_bulk_insert_activity_existing(self):
        """
        WorkerActivity.objects.bulk_insert should steal existing WorkerActivities from other processes
        if those are in a queued or error state
        """
        old_process = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            corpus=self.corpus,
        )
        elements_qs = Element.objects.filter(type__slug='act', type__corpus_id=self.corpus.id)
        self.assertEqual(elements_qs.count(), 5)

        self.process.activities.all().delete()

        # Create one WorkerActivity for each possible state
        WorkerActivity.objects.bulk_create([
            WorkerActivity(
                element=element,
                worker_version=self.worker_version,
                state=state,
                process=old_process,
                configuration=self.configuration,
                started=datetime.now(timezone.utc),
            )
            for element, state in zip(elements_qs, WorkerActivityState)
        ])

        self.assertEqual(self.process.activities.count(), 0)
        self.assertEqual(old_process.activities.count(), 4)
        self.assertCountEqual(list(old_process.activities.values_list('state', flat=True).distinct()), [
            WorkerActivityState.Queued,
            WorkerActivityState.Started,
            WorkerActivityState.Processed,
            WorkerActivityState.Error,
        ])

        with self.assertNumQueries(1):
            WorkerActivity.objects.bulk_insert(
                self.worker_version.id,
                self.process.id,
                self.configuration.id,
                elements_qs,
            )

        # The old process kept the two started and processed activities
        self.assertEqual(old_process.activities.count(), 2)
        self.assertCountEqual(list(old_process.activities.values_list('state', flat=True).distinct()), [
            WorkerActivityState.Started,
            WorkerActivityState.Processed,
        ])

        # The new process gets the two queued and error activities in a started state as requested,
        # and also gets one extra activity for the act that had no activity before
        self.assertEqual(self.process.activities.count(), 3)
        self.assertCountEqual(list(self.process.activities.values_list('state', flat=True).distinct()), [
            WorkerActivityState.Queued,
        ])

    def test_put_activity_user(self):
        """
        Only Ponos task authentication is allowed on UpdateWorkerActivity
        """
        for user in [None, self.user, self.superuser]:
            self.activity.state = WorkerActivityState.Queued
            self.activity.save(update_fields=['state'])
            if user:
                self.client.force_login(user)
            else:
                self.client.logout()
            with self.assertNumQueries(0):
                response = self.client.put(
                    reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                    {
                        'element_id': str(self.element.id),
                        'process_id': str(self.process.id),
                        'state': WorkerActivityState.Started.value,
                    },
                    content_type='application/json',
                )
                self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_put_activity_task(self):
        """
        Ponos task authentication can be used to update a WorkerActivity.
        Tasks can update any WorkerActivity in their process, without any filter on the WorkerVersion,
        since there is no link between a WorkerVersion and a Task yet.
        """

        with self.assertNumQueries(4):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Started)

    def test_put_activity_task_other_process(self):
        """
        Updating a WorkerActivity of another process should be forbidden.
        Unlike the usual HTTP 409, we can raise a HTTP 400 here as the check requires no extra query and can be done
        in the serializer before the UPDATE query that normally updates any WorkerActivity if it exists.
        """
        rev = Revision.objects.first()
        process2 = Process.objects.create(mode=ProcessMode.Repository, creator=self.user, revision=rev)
        process2.save()
        process2.start()
        task = process2.workflow.tasks.first()

        with self.assertNumQueries(2):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
                HTTP_AUTHORIZATION=f'Ponos {task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'process_id': ['Only WorkerActivities for the process of the currently authenticated task can be updated.'],
        })

        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Queued)

    def test_put_activity_wrong_worker_version(self):
        """
        Raises a generic error in case the worker version does not exists because a single SQL request is performed
        The response is a HTTP_409_CONFLICT
        """
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(uuid.uuid4())}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictEqual(response.json(), {
            '__all__': [
                'Either this worker activity does not exist, is assigned to another process, or '
                f"updating the state to '{WorkerActivityState.Started.value}' is forbidden."
            ]
        })

    def test_put_activity_element_unexisting(self):
        """
        Raises a generic error in case no activity exists for this element
        The response is a HTTP_409_CONFLICT
        """
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(uuid.uuid4()),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictEqual(response.json(), {
            '__all__': [
                'Either this worker activity does not exist, is assigned to another process, or '
                f"updating the state to '{WorkerActivityState.Started.value}' is forbidden."
            ]
        })

    def test_put_activity_process_unexisting(self):
        """
        Raises an error in case the process does not exist
        """
        wrong_process_id = uuid.uuid4()
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(wrong_process_id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'process_id': [f'Invalid pk "{wrong_process_id}" - object does not exist.']
        })

    def test_put_activity_allowed_states(self):
        """
        Check each case state update is allowed depending on the actual state of the activity
        """
        queued, started, error, processed = (
            WorkerActivityState[state].value
            for state in ['Queued', 'Started', 'Error', 'Processed']
        )
        allowed_states_update = (
            (queued, started, 4),
            (started, processed, 4),
            (started, error, 5),
            (error, queued, 4),
            (error, started, 4),
        )
        known_date = datetime(2012, 12, 12, 12, 12, 12, 12, timezone.utc)
        for state, payload_state, nb_query in allowed_states_update:
            with self.subTest(previous_state=state, new_state=payload_state):
                self.activity.state = state
                if state == started:
                    self.activity.started = known_date
                else:
                    self.activity.started = None
                self.activity.save()
                with self.assertNumQueries(nb_query):
                    response = self.client.put(
                        reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                        {
                            'element_id': str(self.element.id),
                            'process_id': str(self.process.id),
                            'state': payload_state,
                        },
                        content_type='application/json',
                        HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.activity.refresh_from_db()
                self.assertEqual(self.activity.state.value, payload_state)
                if state == started:
                    # When updating from started to another state, the started time should not be changed at all
                    self.assertEqual(self.activity.started, known_date)
                elif payload_state == started:
                    # When updating to started, we had set a None start timestamp, so it shouldn't be None anymore
                    self.assertIsNotNone(self.activity.started)
                else:
                    # When the started state is not involved, the start timestamp is not set
                    self.assertIsNone(self.activity.started)

    def test_put_activity_forbidden_states(self):
        """
        Check state update is forbidden for some non consistent cases
        The response is a HTTP_409_CONFLICT
        """
        queued, started, error, processed = (
            WorkerActivityState[state].value
            for state in ['Queued', 'Started', 'Error', 'Processed']
        )
        forbidden_states_update = (
            (queued, queued),
            (started, queued),
            (queued, error),
            (processed, error),
        )
        for state, payload_state in forbidden_states_update:
            self.activity.state = state
            if state == started:
                self.activity.started = datetime.now(timezone.utc)
            else:
                self.activity.started = None
            self.activity.save()
            with self.assertNumQueries(4):
                response = self.client.put(
                    reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                    {
                        'element_id': str(self.element.id),
                        'process_id': str(self.process.id),
                        'state': payload_state,
                    },
                    content_type='application/json',
                    HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
                )
                self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
            self.assertDictEqual(response.json(), {
                '__all__': [
                    'Either this worker activity does not exist, is assigned to another process, or '
                    f"updating the state to '{payload_state}' is forbidden."
                ]
            })
            self.activity.refresh_from_db()
            self.assertEqual(self.activity.state.value, state)

    # Just set the timeout to zero so that the transition always works
    @override_settings(WORKER_ACTIVITY_TIMEOUT=0)
    def test_put_activity_retry(self):
        self.activity.state = WorkerActivityState.Started
        self.activity.started = datetime.now(timezone.utc)
        self.activity.save()

        with self.assertNumQueries(4):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
                content_type='application/json',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Started)

    @override_settings(WORKER_ACTIVITY_TIMEOUT=10000)
    def test_put_activity_retry_limit(self):
        self.activity.state = WorkerActivityState.Started
        self.activity.started = datetime.now(timezone.utc)
        self.activity.save()

        with self.assertNumQueries(4):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
                content_type='application/json',
            )
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)

        self.assertDictEqual(response.json(), {
            '__all__': [
                'Either this worker activity does not exist, is assigned to another process, or '
                "updating the state to 'started' is forbidden."
            ]
        })
        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Started)

    def test_put_activity_update_other_activities(self):
        """
        The error update should update the status on the following tasks
        """
        worker_version_2 = WorkerVersion.objects.get(worker__slug='dla')
        worker_version_3 = WorkerVersion.objects.get(worker__slug='worker-gpu')
        worker_version_4 = WorkerVersion.objects.create(
            worker=Repository.objects.first().workers.create(
                name='New version',
                slug='new',
                type=self.worker_type,
            ),
            revision=self.worker_version.revision,
            configuration={},
            state=WorkerVersionState.Available,
            docker_image=self.worker_version.docker_image
        )
        # Create runs run_1 > run_2 > run_3
        run_1 = self.process.worker_runs.first()
        run_2 = self.process.worker_runs.create(
            version=worker_version_2,
            parents=[run_1.id],
        )
        self.process.worker_runs.create(
            version=worker_version_3,
            parents=[run_2.id],
        )

        self.process.worker_runs.create(
            version=worker_version_4,
            parents=[],
        )

        # Create activities for run_2, run_3 and run_4
        activity_2 = self.element.activities.create(
            process=self.process,
            worker_version=worker_version_2,
            state=WorkerActivityState.Processed
        )
        activity_3 = self.element.activities.create(
            process=self.process,
            worker_version=worker_version_3,
            state=WorkerActivityState.Queued
        )
        activity_4 = self.element.activities.create(
            process=self.process,
            worker_version=worker_version_4,
            state=WorkerActivityState.Queued
        )

        self.activity.state = WorkerActivityState.Started
        self.activity.started = datetime.now(timezone.utc)
        self.activity.save()

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Error.value,
                },
                HTTP_AUTHORIZATION=f'Ponos {self.task.token}',
                content_type='application/json',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.activity.refresh_from_db()
        self.assertEqual(self.activity.state, WorkerActivityState.Error)
        activity_2.refresh_from_db()
        self.assertEqual(activity_2.state, WorkerActivityState.Processed)
        activity_3.refresh_from_db()
        self.assertEqual(activity_3.state, WorkerActivityState.Error)
        activity_4.refresh_from_db()
        self.assertEqual(activity_4.state, WorkerActivityState.Error)

    @patch('arkindex.process.models.Process.versions')
    @patch('arkindex.process.models.ActivityManager.bulk_insert')
    def test_async_activities_error(self, bulk_insert_mock, versions_mock):
        """
        Handles errors while initializing workers activity of a process
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='volume'),
            activity_state=ActivityState.Pending
        )
        # Mock list_elements method to return the same queryset object
        elts_qs = Element.objects.filter(type__slug='volume')
        self.assertEqual(elts_qs.count(), 2)
        process.list_elements = MagicMock()
        process.list_elements.return_value = elts_qs

        worker_version_2 = WorkerVersion.objects.get(worker__slug='dla')
        process.worker_runs.create(version=self.worker_version, parents=[], configuration=None)
        process.worker_runs.create(version=worker_version_2, parents=[], configuration=None)

        bulk_insert_mock.side_effect = [
            bulk_insert_mock,
            Exception('Something went wrong')
        ]

        with self.assertRaises(Exception):
            initialize_activity(process)

        self.assertEqual(bulk_insert_mock.call_count, 2)
        self.assertCountEqual(bulk_insert_mock.call_args_list, [
            call(worker_version_id=self.worker_version.id, process_id=process.id, configuration_id=None, elements_qs=elts_qs),
            call(worker_version_id=worker_version_2.id, process_id=process.id, configuration_id=None, elements_qs=elts_qs)
        ])

        process.refresh_from_db()
        self.assertEqual(process.activity_state, ActivityState.Error)
        self.assertEqual(WorkerActivity.objects.filter(element__type__slug='volume').count(), 0)

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_corpus_acl(self):
        private_corpus = Corpus.objects.create(name="can't touch this")
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': private_corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have guest access to this corpus.'})

    def test_list_corpus_not_found(self):
        corpus_id = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'
        self.assertNotEqual(str(self.corpus.id), corpus_id)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': corpus_id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'created': self.activity.created.isoformat().replace('+00:00', 'Z'),
                    'updated': self.activity.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'worker_version_id': str(self.worker_version.id),
                    'configuration_id': str(self.configuration.id),
                    'state': 'queued'
                }
            ]
        })

    def test_list_started(self):
        self.activity.state = WorkerActivityState.Started
        self.activity.started = datetime(2012, 12, 12, 12, 12, 12, 12, timezone.utc)
        self.activity.save()
        # Fetch the updated `updated` timestamp from Postgres, since it is managed by a trigger
        self.activity.refresh_from_db()
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'created': self.activity.created.isoformat().replace('+00:00', 'Z'),
                    'updated': self.activity.updated.isoformat().replace('+00:00', 'Z'),
                    'started': '2012-12-12T12:12:12.000012Z',
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'worker_version_id': str(self.worker_version.id),
                    'configuration_id': str(self.configuration.id),
                    'state': 'started',
                }
            ]
        })

    def test_list_filter_process_wrong_corpus(self):
        private_corpus = Corpus.objects.create(name="can't touch this")
        process = private_corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=self.superuser,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}),
                {'process_id': str(process.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'process_id': ['This process does not exist.']})

    def test_list_filter_process(self):
        process = self.corpus.processes.create(
            mode=ProcessMode.Workers,
            creator=self.user,
        )
        activity = process.activities.create(
            element=self.element,
            worker_version=self.worker_version,
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(9):
            response = self.client.get(
                reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}),
                {'process_id': str(process.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'created': activity.created.isoformat().replace('+00:00', 'Z'),
                    'updated': activity.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'element_id': str(self.element.id),
                    'process_id': str(process.id),
                    'worker_version_id': str(self.worker_version.id),
                    'configuration_id': None,
                    'state': 'queued'
                }
            ]
        })

    def test_list_filter_state(self):
        element = Element.objects.get(name='Volume 1')
        activity = element.activities.create(
            process=self.process,
            worker_version=self.worker_version,
            state=WorkerActivityState.Processed,
        )
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Processed).count(), 1)
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}),
                {'state': 'processed'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'previous': None,
            'next': None,
            'results': [
                {
                    'created': activity.created.isoformat().replace('+00:00', 'Z'),
                    'updated': activity.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'element_id': str(element.id),
                    'process_id': str(self.process.id),
                    'worker_version_id': str(self.worker_version.id),
                    'configuration_id': None,
                    'state': 'processed'
                }
            ]
        })

    def test_list_invalid_filters(self):
        self.client.force_login(self.superuser)
        cases = [
            (
                {'process_id': 'a'},
                {'process_id': ['Process ID should be an UUID.']},
            ),
            (
                {'process_id': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'},
                {'process_id': ['This process does not exist.']},
            ),
            (
                {'state': 'lol'},
                {'state': ["'lol' is not a valid WorkerActivity state"]},
            ),
            (
                {'process_id': 'a', 'state': 'lol'},
                {
                    'process_id': ['Process ID should be an UUID.'],
                    'state': ["'lol' is not a valid WorkerActivity state"],
                },
            ),
        ]
        for filters, expected_errors in cases:
            with self.subTest(**filters):
                response = self.client.get(
                    reverse('api:corpus-activity', kwargs={'corpus': self.corpus.id}),
                    filters
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)
