import logging
import uuid
from hashlib import sha256

from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.postgres.fields import ArrayField
from django.core.exceptions import ValidationError
from django.core.validators import MinLengthValidator
from django.db import models
from django.db.models import Q
from django.utils.functional import cached_property
from enumfields import Enum, EnumField

from arkindex.project.aws import S3FileMixin
from arkindex.project.fields import MD5HashField
from arkindex.project.models import IndexableModel

logger = logging.getLogger(__name__)


class Model(IndexableModel):
    """
    An evolving Machine Learning model
    """
    # Name of the model, unique
    name = models.CharField(max_length=100, unique=True)

    description = models.TextField(default="")

    public = models.BooleanField(default=False)

    # Link to the workers that are able to use this model
    compatible_workers = models.ManyToManyField('process.Worker', related_name='models')

    # Memberships to handle access rights on models
    memberships = GenericRelation('users.Right', 'content_id')

    def __str__(self):
        return self.name


class ModelVersionState(Enum):
    """
    State of the model Version, available meaning checked by the backend
    """
    Created = 'created'
    Available = 'available'
    Error = 'error'


class ModelVersion(S3FileMixin, IndexableModel):
    """
    A specific Model version
    """
    model = models.ForeignKey('training.Model', related_name='versions', on_delete=models.CASCADE)

    parent = models.ForeignKey('self', related_name='children', null=True, blank=True, on_delete=models.CASCADE)

    description = models.TextField(default="")

    tag = models.CharField(null=True, max_length=50, default=None)

    state = EnumField(ModelVersionState, default=ModelVersionState.Created)

    # Hash of the archive's content
    hash = MD5HashField(
        null=True,
        blank=True,
        help_text="hash of the content of the archive which contains the model version's data",
    )

    # Hash of the archive
    archive_hash = MD5HashField(
        null=True,
        blank=True,
        help_text="hash of the archive which contains the model version's data",
    )

    # Size of the archive
    size = models.PositiveIntegerField(null=True, blank=True, help_text='file size in bytes')

    # Store dictionary of paramseters given by the ML developer
    configuration = models.JSONField(default=dict)

    s3_bucket = settings.AWS_TRAINING_BUCKET

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['model', 'tag'],
                name='modelversion_unique_tag',
            ),
            models.UniqueConstraint(
                fields=['model', 'hash'],
                name='modelversion_unique_hash',
                condition=Q(hash__isnull=False),
            ),
        ]

    @cached_property
    def s3_key(self):
        return f'{self.id}.zst'

    @property
    def truncated_id(self):
        return str(self.id)[0:6]

    def perform_check(self, save=True, raise_exc=False):
        """
        Check ModelVersion's existence, content type and size from S3.
        Have to override this method because it uses a different enum with different statuses
        """
        try:
            assert self.exists(), 'Archive has not been uploaded'
            assert self.s3_object.content_length == self.size, \
                f'Uploaded file size is {self.s3_object.content_length} bytes, expected {self.size} bytes'
            self.state = ModelVersionState.Available
        except AssertionError:
            self.state = ModelVersionState.Error
            if raise_exc:
                raise
        finally:
            if save:
                self.save()

    def check_hash(self, save=True, raise_exc=False):
        """
        Checks the MD5 hash against the hash from Amazon S3
        """
        assert self.archive_hash, 'File has no hash'
        assert self.exists(), 'No file content, assert file has been correctly uploaded'
        # The hash given by Boto seems to be surrounded by double quotes
        if self.s3_object.e_tag.strip('"') == self.archive_hash:
            self.state = ModelVersionState.Available
        elif '-' in self.s3_object.e_tag:
            # Multipart hash: a hash of each part's hash,
            # combined with the number of parts, separated by a dash
            logger.warning('Could not check remote multipart hash {!r} against local hash {!r}'.format(
                self.s3_object.e_tag,
                self.archive_hash,
            ))
            self.state = ModelVersionState.Available
        else:
            self.state = ModelVersionState.Error
        if save:
            self.save()
        if self.state == ModelVersionState.Error and raise_exc:
            raise ValueError('MD5 hashes do not match')

    def build_authentication_token_hash(self):
        """
        This allows to authenticate the caller when downloading a model version without requiring standard API authentication.
        """
        return sha256((str(self.id) + self.hash + settings.SECRET_KEY).encode('utf-8')).hexdigest()


class MetricMode(Enum):
    Series = 'series'
    Point = 'point'


class MetricKey(models.Model):
    """
    A metric generated by a model training
    """
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    name = models.CharField(max_length=100)
    model_version = models.ForeignKey('training.ModelVersion', related_name='metrics', on_delete=models.CASCADE)
    mode = EnumField(MetricMode, default=MetricMode.Series)

    class Meta:
        unique_together = (
            ('name', 'model_version')
        )


class MetricValue(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    metric = models.ForeignKey('training.MetricKey', related_name='values', on_delete=models.CASCADE)
    value = models.FloatField()
    created = models.DateTimeField(auto_now_add=True)
    step = models.PositiveIntegerField(blank=True, null=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['metric', 'step'], name='metric_unique_step', condition=Q(step__isnull=False)),
            models.UniqueConstraint(fields=['metric', 'created'], name='metric_unique_no_step', condition=Q(step__isnull=True)),
        ]


class DatasetState(Enum):
    Open = 'open'
    Building = 'building'
    Complete = 'complete'
    Error = 'error'


def validate_unique_set_names(sets):
    if len(set(sets)) != len(sets):
        raise ValidationError('Set names must be unique.')


def default_sets():
    return ['training', 'test', 'validation']


class Dataset(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    corpus = models.ForeignKey(
        'documents.Corpus',
        related_name='datasets',
        on_delete=models.DO_NOTHING,
    )
    creator = models.ForeignKey(
        'users.User',
        related_name='datasets',
        on_delete=models.DO_NOTHING,
    )
    task = models.ForeignKey(
        'ponos.Task',
        related_name='dataset',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    name = models.CharField(max_length=100, validators=[MinLengthValidator(1)])
    description = models.TextField(validators=[MinLengthValidator(1)])
    state = EnumField(DatasetState, default=DatasetState.Open)

    sets = ArrayField(
        models.CharField(max_length=50, validators=[MinLengthValidator(1)]),
        validators=[
            MinLengthValidator(1),
            validate_unique_set_names,
        ],
        default=default_sets,
    )

    elements = models.ManyToManyField(
        'documents.Element',
        through='training.DatasetElement',
        related_name='datasets',
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['corpus', 'name'],
                name='unique_dataset_names',
            ),
        ]

    def __str__(self):
        return self.name


class DatasetElement(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True, editable=False)
    dataset = models.ForeignKey(
        Dataset,
        related_name='dataset_elements',
        on_delete=models.DO_NOTHING,
    )
    element = models.ForeignKey(
        'documents.Element',
        related_name='dataset_elements',
        on_delete=models.DO_NOTHING,
    )
    set = models.CharField(max_length=50, validators=[MinLengthValidator(1)])

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['dataset', 'element', 'set'],
                name='unique_dataset_elements',
            ),
        ]
